// Copyright (c) 2014, <Richard Ambler>. All rights reserved. Use of this 
// source code is governed by a BSD-style license that can be found in the 
// LICENSE file.

import 'dart:html';
import 'package:fisher/fisher.dart';

// Import the iris data set.
import 'package:fisher/data_sets/iris.dart';

void main() {
  var 
    // Create a new plot object.
    plot1 = new Plot(), 
    x = iris["Petal.Width"], 
    y = iris["Petal.Length"],
    // Generate a simple linear model by species.
    models = new Map.fromIterable(
      iris["Species"].categories,
      value: (species) => Model.simpleLinearModel(
        x[iris["Species"] == species], 
        y[iris["Species"] == species]
      )
    )
  ;
  
  // Add the model that ignores species.
  models["all"] = Model.simpleLinearModel(x, y);
  
  // Choose some colours for the species.
  var colors = new Map.fromIterables(
    iris["Species"].categories, 
    [
       "rgba(25, 125, 50, 0.5)", 
       "rgba(200, 100, 25, 0.5)", 
       "rgba(75, 25, 125, 0.5)"
    ]
  );
  
  // Add a blue for the overall model.
  colors["all"] = "rgba(0, 0, 255, 0.5)";
  
  // Create a vector that may be passed to the colors argument.
  var colorsVec = iris["Species"].map((x) => colors[x]);
  
  plot1
    ..scatter(
        x: x, 
        y: y, 
        plotPoints: false,
        xLabel: "Width",
        yLabel: "Length",
        title: "Iris Petals"
      )
    ..points(x: x, y: y, fillColor: colorsVec)
  ;
  
  // Draw a key.
  plot1.polygon(
    x: [-0.05, -0.05, 0.7, 0.7], 
    y: [5.65, 7.15, 7.15, 5.65], 
    fillColor: "rgba(255,255,255,0.6)", 
    lineColor: "rgba(0, 0, 0, 0.6)"
  );
  int i = 0; models.forEach((key, model) {
    plot1
      ..curve(model: model, color: colors[key])
      ..labels(
          x: 0.2, 
          y: 6 + i / 4, 
          align: "left", 
          labels: key,
          color: colors[key],
          font: "Arial"
        )
      ..lines(x: [0, 0.15], y: 6 + i / 4, color: colors[key])
    ;
    i++;
  });
  
  // Add the plot to the document.
  document.body.children.add(plot1.canvas);
  
  // Take a look at the residuals
  var plot2 = new Plot();
  plot2.histogram(
    data: models["all"].info["residuals"],
    xLabel: "Residual",
    yLabel: "Frequency"
  );
  document.body.children.add(plot2.canvas);
  print(models["all"].summary);
}