part of fisher;

/// The base vector class.
abstract class _Vector {
  List _data;
  List get data;
/// The name of the vector.
  String name;
  
/// The number of elements in the vector.
///
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     print(a.count);
/// 
/// ...
  int get count => _data.length;
  
/// Returns a Vector containing only the elements at indices [index].
/// 
/// Expects [index] to be of type [int], [List] or [BooleanVector].
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     var b = a.takeIndices([0, 4, 5]);
/// 
/// ...
  takeIndices(var index) => // overridden for categoricVector
    // return the actual value if the index is an int
    index is int ? _data[index % count] : 
      index is BooleanVector ? takeIndices(index.indicesWhere(true)) :
        _vector(
           this,
           new List.generate(
             index.length, 
             (int i) => _data[index[i] % count]
           ),
           name: name
         );
  
  operator [] (index) => takeIndices(index);
  
/// Returns a vector containing the last few instances.
  tail({int take: 10}) {
    int min = Math.min(take, count);
    return this[
      new List<int>.generate(
        min,
        (int i) => - (min - i)
      )
    ];
  }
  
/// Returns a vector containing the first few instances.
  head({int take: 10}) => this[
    new List<int>.generate(
      Math.min(take, count),
      (int i) => i
    )
  ];

/// Returns a vector obtained from mapping the elements under [f].
  map(Function f) => vector(data.map(f).toList());

/// Returns a vector of those elements for which [f] returns true.
  where(Function f) => vector(data.where(f).toList());  
  
/// Returns whether [test] returns true for any of the elements.
  bool any(Function test) => data.any(test);
  
/// Returns whether [test] returns true for every one of the elements.
  bool every(Function test) => data.every(test);
  
  @override String toString() => "$name (${this.runtimeType}) $_data";
  
}
