part of fisher;

/// An indexable Boolean vector.
class BooleanVector extends _Vector {
  @override List<bool> _data;
  
/// The instances stored in the vector.
  @override List<bool> get data => new List<bool>.from(_data);
  
  BooleanVector({data, String name}) {
    this.name = name == null ? DEFAULT_VECTOR_NAME : name;
    _data = new List<bool>.from(data);
  }
  
/// Returns the number of elements that have value [x].
/// 
/// ...
/// Example:
/// 
///     var a = new BooleanVector([true, true, false, true, false]);
///     print(a.countif(false));
/// 
/// ...
  int countIf(bool x) => 
    _data
      .map((bool a) => a == x ? 1 : 0)
      .fold(0, (int a, int b) => a + b);
  
/// The number of positives (true) in the vector.
  int get sum => countIf(true);
  
/// Returns returns a [List] containing the indices of
/// elements that have the value [x].
/// 
/// ...
/// Example:
///
///     var a = new BooleanVector([true, true, false, true, false]);
///     print(a.indicesWhere(false));
/// 
/// ...
  List<int> indicesWhere(bool x) =>
    new List<int>
      .generate(count, (int i) => _data[i] ? i : -1)
      .where((int i) => i != -1)
      .toList();
  
  _booleanOperation(List<bool> x, Function f) {
    return new BooleanVector(
      data: new List<bool>.generate(
        count, 
        (int i) => f(_data[i], x[i % x.length])
      ),
      name: name
    );
  }
  
/// Performs element-wise conjunction. 
/// 
/// Expects [x] to be a [BooleanVector], a [List] or a [bool].
/// 
  operator & (var x) {
    List<bool> xList = _extractBoolList(x);
    return _booleanOperation(xList, (bool a, bool b) => a && b);
  }
  
/// Performs element-wise disjunction. 
///
/// Expects [x] to be a [BooleanVector], a [List] or a [bool].
/// 
  operator | (var x) {
    List<bool> xList = _extractBoolList(x);
    return _booleanOperation(xList, (bool a, bool b) => a || b);
  }
  
/// Performs element-wise exclusive disjunction. 
///
/// Expects [x] to be a [BooleanVector], a [List] or a [bool].
///
  operator ^ (var x) {
    List<bool> xList = _extractBoolList(x);
    return _booleanOperation(xList, (bool a, bool b) => (a || b) && !(a && b));
  } 
  
/// Performs element-wise exclusive equivalence. 
///
/// Expects [x] to be a [BooleanVector], a [List] or a [bool].
///
  @override operator == (var x) {
    List<bool> xList = _extractBoolList(x);
    return _booleanOperation(xList, (bool a, bool b) => a == b);
  }
}
