part of fisher;

/// An indexable vector of category instances.
class CategoricVector extends _Vector {
  static String NO_CATEGORY = "<NO CATEGORY>";  

  @override List<int> _data;
  List<String> categories;

/// The instances stored in the vector by category name.
  @override List<String> get data => new List<String>.generate(
    _data.length,
    (int i) => i == -1 ? NO_CATEGORY : categories[_data[i]]
  );

/// The instances stored in the vector by category index.
  List<int> get dataAsInts => new List<int>.from(_data);
  
  CategoricVector({List data, String name, List<String> categories}) {
    this
      ..name = name == null ? DEFAULT_VECTOR_NAME : name
      ..categories = categories == null ? 
        (data.map((x) => x.toString()).toList().toSet().toList()..sort()) :
        categories
    ;  
    _data = new List<int>.generate(
      data.length, 
      (int i) => this.categories.indexOf(data[i].toString())
    );
  }
  
/// Returns a Vector containing the elements at [index].
/// 
/// Expects [index] to be of type [int], [List] or [BooleanVector].
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     var b = a.takeIndices([0, 4, 5]);
///     
/// ...
  @override takeIndices(var index) =>
    index is int ? (_data[index % count] == -1 ? NO_CATEGORY :
      categories[_data[index % count]]) : // takeIndices([index]) :
      index is BooleanVector ? takeIndices(index.indicesWhere(true)) :
        _vector(
          this,
          new List.generate(
            index.length, 
            (int i) => _data[index[i] % count] == -1 ? NO_CATEGORY :
              categories[_data[index[i] % count]]
          ),
          name: name,
          categories: categories
        );
   
  @override String toString() => 
    "$name (${
      this.runtimeType})\n  Categories: $categories\n  ${
        _data.map((x) => x == -1 ? NO_CATEGORY : categories[x]).toList()}";
              
  @override 
    operator == (var x) => 
      x is String ? this == [x] :
        x is CategoricVector ? this == x.data :
          new BooleanVector(
            data: new List<bool>.generate(
              _data.length, 
              (int i) => this[i] == x[i % x.length]
            ),
            name: name
          );
 
/// Returns the counts of the instances by category as a [Map].
  Map<String, int> get countByCategory =>
    new Map.fromIterable(
      categories,
      value: (String x) => ((this == x) as BooleanVector).sum
    );
}
