part of fisher;

/// An indexable represantation of a data frame or matrix.
class DataMatrix {
  
  Map<String, _Vector> _data;
  Map _cache = {};


// TODO: Implement weights
  List<num> _weights = [1];
  /// The weights associated with the instances.
  List<num> get weights => _weights;
  set weights (var x) {
    _weights = _extractNumList(x);
    for (String name in names) 
      if (_data[name] is NumericVector) 
        (_data[name] as NumericVector).clearCache();
  }
  
  DataMatrix([List<_Vector> data]) {
    if (data != null) {
      int i = 0;
      _data = new Map<String, _Vector>.fromIterable(
        data, 
        key: (_Vector v) { 
          if (v.name == null || v.name == DEFAULT_VECTOR_NAME) {
            String name = "C$i";
            i++;
            return name;
          } else return v.name;
        },
        value: (_Vector v) => v
      );
    } else {
      _data = new Map<String, _Vector>();
    }
  }
  
/// Constructs a [DataMatrix] from a csv representation.
/// 
/// Expects [csv] to be a [String] or a [List].
  DataMatrix.fromCSV(var csv, {String separator: ",", bool header: true}) {
    
    Function helper = (List<String> lines) {
      List<String> names = header ? 
        lines[0].split(separator) : 
          new List<String>.generate(
            lines[0].split(separator).length, 
            (int i) => "C$i"
          );
      List<List> data = new List.generate(names.length, (_) => new List());
      
      for (int i = header ? 1 : 0; i < lines.length; i++) {
        List datum = lines[i].split(separator);
        for (int j = 0; j < datum.length; j++) 
          data[j].add(datum[j]);
        
      }
      
      for (int i = 0; i < data.length; i++) {
        try {
          data[i] = data[i].map((String s) => int.parse(s)).toList();
        } on FormatException {
          try {
            data[i] = data[i].map((String s) => double.parse(s)).toList();
          } on FormatException {
            if (data[i].every((String s) => s == "true" || s == "false")) 
              data[i] = data[i].map((String s) => s == "true").toList();
          }
        }
        _data[names.elementAt(i)] = vector(data[i]);
        _data[names.elementAt(i)].name = names[i];
      }
                
    };
    
    _data = new Map<String, _Vector>();
    helper(csv is String ? csv.split("\n") : csv);
  }
 
/// Constructs a [DataMatrix] from a [List] of [List]s.
  DataMatrix.fromLists(List<List<num>> m) {
    int r = m.length, c = m.fold(0, (a, List b) => Math.max(a, b.length));
    _data = new Map<String, _Vector>();
    
    for (int i = 0; i < c; i++) 
      _data["C$i"] = new NumericVector(data: new List.filled(r, 0)); 
    
    for (int i = 0; i < r; i++) setRow(row: i, value: m[i]);
  }
  
/// The number of rows. 
/// 
/// If the column variables have different [count]s, they are reused 
/// so that the data may be interpreted as a rectangular matrix, in 
/// which case [rows] represents the maximum count.
  int get rows => _cache.containsKey("rows") ? _cache["rows"] : 
    _cache["rows"] = new List<int>.generate(
      _data.length, 
      (int i) => _data[_data.keys.elementAt(i)].count
    ).reduce((int value, int i) => Math.max(value,  i));
  
/// The number of columns.
  int get cols => _data.length;
  
/// The names of the stored variables stored.
  List<String> get names => _data.keys.toList();

  
// TODO: define operations for arguments of type num...
  
/// Performs matrix multiplication.
/// 
/// Expects [x] to be a [NumericVector], in which case [x] is treated
/// as a column-matrix, or a [DataMatrix].
/// 
/// Returns null if the operation is not possible.
  operator * (var x) {
    if (x is NumericVector) return this.matrix * new DataMatrix([x]);
    else {
      DataMatrix m1 = this.matrix, m2 = (x as DataMatrix).matrix;
      if (m1.cols == m2.rows) {
        return new DataMatrix(
          new List<NumericVector>.generate(
            m2.cols, 
            (int i) => new NumericVector(
                data: new List<num>.generate(
                  m1.rows, 
                  (int j) => (new NumericVector(
                    data: new List<num>.generate(
                      m1.cols, 
                      (int k) => m1.datumAt(row: j, col: k)
                    )
                  ).dot(m2[m2._data.keys.elementAt(i)])
                )
              )
            )
          )    
        );
      } 
    }
  }
  
/// Performs matrix addition.
/// 
/// Expects [x] to be a [NumericVector], in which case [x] is treated
/// as a column-matrix, or a [DataMatrix].
/// 
/// Returns null if the operation is not possible.  
  operator + (x) {
    if (x is NumericVector) return this.matrix + new DataMatrix([x]);
    else {
      DataMatrix m1 = this.matrix, m2 = (x as DataMatrix).matrix;
      if (m1.rows == m2.rows && m1.cols == m2.cols) {
        return new DataMatrix(
          new List<NumericVector>.generate(
            m1.cols, 
            (int i) => new NumericVector(
              data: m1[m1.names[i]] + m2[m2.names[i]]
            )
          )    
        );
      }
    }
  }

/// Performs matrix subtraction.
/// 
/// Expects [x] to be a [NumericVector], in which case [x] is treated
/// as a column-matrix, or a [DataMatrix].
/// 
/// Returns null if the operation is not possible.
  operator - (x) {
    if (x is NumericVector) return this.matrix + new DataMatrix([x]);
    else {
      DataMatrix m1 = this.matrix, m2 = (x as DataMatrix).matrix;
      if (m1.rows == m2.rows && m1.cols == m2.cols) {
        return new DataMatrix(
          new List<NumericVector>.generate(
            m1.cols, 
            (int i) => new NumericVector(
              data: m1[m1.names[i]] - m2[m2.names[i]]
            )
          )    
        );
      }
    }
  }

/// Performs matrix negation.
  operator - () {
    var m = matrix;
    return new DataMatrix(
      new List<NumericVector>.generate(
        m.cols, 
        (int i) => new NumericVector(
           data: -m[m.names[i]]
        )
      )    
    );
  }
      
/// Returns the [r]th row.
/// 
/// Non numeric columns are ignored. 
  NumericVector row(int r) {
    List<String> numericCols = _data.keys.where(
      (String name) => _data[name] is NumericVector).toList();
    return new NumericVector(
      data: new List<num>.generate(
        numericCols.length, 
        (int i) => _data[numericCols[i]][r]
      )    
    );
  }

/// Sets the values in the [r]th row.
/// 
/// Non numeric columns are ignored. Values are reused if necessary. 
  void setRow({int row, value}) {
    if (value is num) {
      setRow(row: row, value: [value]);
    } else if (value is NumericVector) {
      setRow(row: row, value: value.data);
    } else {
      // Assume x is List<num>
      int index = 0;
      for (String name in _data.keys) {
        if (_data[name] is NumericVector) {
          (_data[name] as NumericVector)[row] = value[index % value.length];
          index++;
        }
      }
    }
  }

/// Swaps rows [r1] and [r2].
  void swapRows(int r1, int r2) {
    NumericVector temp = row(r1);
    setRow(row: r1, value: row(r2));
    setRow(row: r2, value: temp);
  }

/// Attempts to solve a matrix equation.
/// 
/// Returns null if unable to solve.
/// 
/// ...
/// Example:
/// 
///     var
///       A = matrix([
///         [1,  2,  3],
///         [2, -1,  1],
///         [3,  0, -1]
///       ]),
///       B = matrix([
///         [9, 8, 3]
///       ]).transpose
///     ;
///    
///     print(A.solve(B));
///  
/// ...     
  DataMatrix solve(DataMatrix target) {
    DataMatrix A = matrix, B = target.matrix;
    for (int pivot = 0; pivot < A.rows; pivot++) {
      int r;
      for (r = pivot; r < A.rows; r++) if (A.row(r)[pivot] != 0) break;
      if (r == A.rows) return null;
      if (r != pivot) {
        A.swapRows(r, pivot);
        B.swapRows(r,  pivot);
      }
      num d = A.datumAt(row: pivot, col: pivot);
    
      A.setRow(row: pivot, value: A.row(pivot) / d);
      B.setRow(row: pivot, value: B.row(pivot) / d);
      
      for (int r = 0; r < A.rows; r++) {
        if (r == pivot) continue;
        num d2 = A.row(r)[pivot];
        A.setRow(row: r, value: A.row(r) - A.row(pivot) * d2);
        B.setRow(row: r, value: B.row(r) - B.row(pivot) * d2);
      }
    }
    
    return B;
  }

/// The inverse of the matrix.
/// 
/// Returns null if unable to solve.
/// 
/// ...
/// Example:
/// 
///     var A = matrix([
///       [1,  2,  3],
///       [2, -1,  1],
///       [3,  0, -1]
///     ]);
///   
///     print(A.inverse);
///     
/// ...
  DataMatrix get inverse {
    DataMatrix A = matrix;
    if (A.rows != A.cols) return null;
    else return A.solve(I(A.rows));
  }
  
/// Returns the [c]th column if it is of type [NumericVector], otherwise null. 
  NumericVector col(int c) {
    var x = _data[_data.keys.elementAt(c)];
    if (x is NumericVector) 
      return new NumericVector(data: x.data, name: x.name);
    else 
      return null;
  }
  
  operator [] (x) =>
    x is String ? _data[x] :
      x is List && x.first is String ? 
          new DataMatrix(
            x.map((String name) => _data[name]).toList()    
          ):
      x is BooleanVector ? this[x.indicesWhere(true)] :
        x is int ? row(x) : //this[[x]] :
          
          // Assume x is List<int>
          new DataMatrix(
            new List<_Vector>.generate(
              _data.length, 
              (int i) => _data[_data.keys.elementAt(i)][x]
            )  
          );
  
  operator []= (String x, _Vector value) {
    _data[x] = value;
    _data[x].name = x;
  }
  
/// A [List] containing the names of the numeric variables.
  List<String> get numerics => _data.keys.where(
    (String name) => _data[name] is NumericVector).toList(); 
  
  
/// A [DataMatrix] containing only the numeric variables.
  DataMatrix get matrix => new DataMatrix(
    numerics.map((String name) => 
      new NumericVector(data: _data[name])).toList()   
  );
  
/// The transpose. 
/// 
/// Non numeric vectors are ignored.
  DataMatrix get transpose => new DataMatrix(
      new List<NumericVector>.generate(
        rows,
        (int r) => row(r)
      )    
    );

  
/// Returns the value at a specific row and column.
/// 
/// Expects [col] to be a [String] or an [int].
  datumAt({int row, var col}) =>
    col is String ? _data[col][row] :
      _data[_data.keys.elementAt(col)][row];
  
/// Returns a [DataMatrix] containing only the last few rows.
  DataMatrix tail({int take: 10}) {
    int min = Math.min(take, rows);
    return this[
      new List<int>.generate(
        min,
        (int i) => - (min - i)
      )
    ];
  }
  
/// Returns a [DataMatrix] containing only the first few rows.
  DataMatrix head({int take: 10}) => this[
    new List<int>.generate(
      Math.min(take, rows),
      (int i) => i
    )
  ];
  
  @override String toString() {
    int padding = 10;
    for (String name in _data.keys) 
      if (name.length > padding) padding = name.length;
    padding += 1;     
    var sb = new StringBuffer();
    var writeItem = (x) {
      if (x is num) {
        num p = Math.pow(10, DIGITS);
        x = (x * p).round() / p;
      }
      String s = x.toString();
      sb.writeAll([
        " " * (padding - s.length),
        s
      ]);
    };
      
    sb.write("${this.runtimeType}\n");
    for (String name in _data.keys) writeItem(name);
    sb.write("\n");
     
          
    int start = 0, stop = rows;
    for (int i = start; i < stop; i++) {
      for (String name in _data.keys) writeItem(_data[name][i]);
      sb.write("\n");
    }
    return sb.toString();
  }
  
/// Returns a [String] csv representation of the stored data.
  String toCSV({String separator: ",", bool includeNames: true}) {
    var sb = new StringBuffer();
    if (includeNames) {
      sb.write("[index]$separator");
      sb.write(_data.keys.join(separator));
      sb.write("\n");
    }
    for (int i = 0; i < rows; i++) {
      List<String> line = ["$i"];
      for (String variable in _data.keys)  
        line.add(_data[variable][i].toString());
      sb.write(line.join(separator));
      sb.write("\n");
    }
    return sb.toString();
  }
   
/// Returns a [String] markdown representation of the stored data.
  String get markDown {
    var sb = new StringBuffer();
    sb.writeAll([
      "|[index]|",
      _data.keys.join("|"),
      "|\n",
      "|:-:|",
      _data.keys.map((_) => ":-:").join("|"),
      "|:-:|\n"
    ]);
    
    for (int i = 0; i < rows; i++) {
      List<String> line = ["$i"];
      for (String variable in _data.keys)  
        line.add(_data[variable][i].toString());
      sb.writeAll([
        "|",
        line.join("|"),
        "|\n"
      ]);
    }
    return sb.toString();
  }
  
/// Returns a [String] JSON representation of the stored data.
 String get JSON {
    var sb = new StringBuffer();
    sb.write("{\n");
    int i = _data.length;
    _data.forEach((String name, value) {
      sb.write("  \"$name\": ");
      if (value is CategoricVector) 
        sb.writeAll([
          "[",
          value.data.map((String v) => "\"$v\"").join(", "),
          "]"
        ]);
      else sb.write(value.data);
      i--;
      if (i > 0) sb.write(",");
      sb.write("\n");
    });
    sb.write("}\n");
    
    return sb.toString();
  }
 
// TODO: clean up Future use...
 
/// Returns a [Plot] containing summary cross-plots of the data.
 Plot summaryPlot({
   int width: 800,
   int height
 }) {
   var boolToCat = (String name) => new CategoricVector(
     data: _data[name].map((bool x) => x.toString()));
   num margin = width ~/ 10;
   if (height == null) height = width;
   Plot p = new Plot(width: width + 2 * margin, height: height + 2 * margin);
   for (int i = 0; i < _data.keys.length; i++) {
     String name = _data.keys.elementAt(i);
     p.context.font = "${width ~/ 65}px Arial Rounded MT Bold";
     num halfWidth = p.context.measureText(name).width / 2;
     p.context
       ..fillStyle = "gray"
       ..fillText(
           name, 
           2 * margin + i * width / _data.keys.length - halfWidth, 
           margin * 0.75
         )
       ..fillText(
           name, 
           2 * margin + i * width / _data.keys.length - halfWidth, 
           height + margin * 1.4
         )
       ..save()
       ..translate(
           margin * 0.75, 
           2 * margin + i * height / _data.keys.length + halfWidth
         )
       ..rotate(-Math.PI / 2)
       ..fillText(
           name, 
           0, 
           0
         )
       ..restore()
       ..save()
       ..translate(
           width + margin * 1.2, 
           2 * margin + i * height / _data.keys.length - halfWidth
         )
       ..rotate(Math.PI / 2)
       ..fillText(
           name, 
           0, 
           0
         )
       ..restore()
     ;
     
     var xVec = _data[name] is BooleanVector ? 
       boolToCat(name) : 
         _data[name];
     
     for (int j = 0; j < _data.keys.length; j++) {
       String y = _data.keys.elementAt(j);
       var yVec = _data[y] is BooleanVector ? 
         boolToCat(y) : 
           _data[y];
       
       Plot sub = new Plot(
         width: width ~/ _data.keys.length, 
         height: height ~/ _data.keys.length
       );
       
       new Future(() {
         if (name == y) {
           if (_data[name] is NumericVector) {
             sub.histogram(
               data: _data[name], 
               summary: true
             );
           } else if (_data[name] is CategoricVector) {
             sub.barChart(
               categories: _data[name], 
               summary: true
             );
           }
         }
         else if (_data[name] is NumericVector) {
           if (_data[y] is NumericVector) {
             sub.scatter(
               x: _data[name], 
               y: _data[y], 
               summary: true
             );
           } else if (_data[y] is CategoricVector) {
             sub.boxPlot(
               data: _data[name], 
               categories: _data[y], 
               horizontal: true, 
               summary: true
             );  
           }
         } else if (_data[name] is CategoricVector) {
           if (_data[y] is NumericVector) {
              sub.boxPlot(
                data: _data[y], 
                categories: _data[name], 
                summary: true
              );   
            } else if (_data[y] is CategoricVector) {
              sub.mosaic(
                categoriesX: _data[name], 
                categoriesY: _data[y], 
                summary: true
              );         
            }
          } 
         
          p.context.drawImage(
            sub.canvas, 
            i * sub.canvas.width + margin, 
            j * sub.canvas.height + margin
          );
        });
      }
     
    }
    return p;
  }
}