part of fisher;

/// Returns an arithmetic sequence of numbers.
/// 
///  The sequence is returned as a [List] of numbers that starts at [from] 
///  and ends at [until]. The common difference can be specified by [by] 
///  or inferred from the number of elements in the returned [List]
/// 
/// ...
/// Example:
/// 
///     print(sequence(0, 10, by: 2));
/// 
/// ...
List<num> sequence(
    num from, 
    num until, 
    {
      num by, 
      num length,
      bool inclusive: false
    }
) {
  if (length != null) by = (until - from) / length;
  if (by == null) by = from > until ? -1: 1;
  return new List<num>.generate(
    ((until - from) / by).ceil() + (inclusive ? 1 : 0),
    (int i) => from + i * by
  );
}

/// Returns a vector of type determined from [data].
/// 
/// ...
/// Example:
/// 
///     var a = vector([5, 4, 6, 3, 7, 2]);
///     print(a);
///     
/// ...
vector(List data, {String name}) =>
  data.first is bool ? new BooleanVector(data: data, name: name) :
    data.first is num ? new NumericVector(data: data, name: name) :
      new CategoricVector(data: data, name: name);

/// Returns a [DataMatrix] from a [List] of [List]s.
/// 
/// ...
/// Example:
/// 
///     var A = matrix([
///       [1,  2,  3],
///       [2, -1,  1],
///       [3,  0, -1]
///     ]);
///     
/// ...
DataMatrix matrix(List<List> data) => 
  new DataMatrix.fromLists(data);

/// Returns an [n] by [n] identity matrix 
DataMatrix I(int n) => new DataMatrix(
  new List<_Vector>.generate(
    n, 
    (int i) => new NumericVector(
      data:new List<num>.generate(
        n,
        (int j) => i == j ? 1 : 0
      )
    )
  )
);