part of fisher;

/// An abstract class to facilitate model construction.
abstract class Model {
  static SimpleLinearModel simpleLinearModel(x, y) =>
    new SimpleLinearModel.fromData(x, y);
  
}

abstract class _Model {
  Function f;
  Map info = {};
  
  predictFor(x) => f(x);
}

/// A simple linear model.
class SimpleLinearModel extends _Model {
  SimpleLinearModel();
  SimpleLinearModel.fromData(x, y) {
    var 
      xVec = x is NumericVector ? x :
        new NumericVector(data: _extractNumList(x), name: "X"),
      yVec = y is NumericVector ? y :
        new NumericVector(data: _extractNumList(y), name: "Y")
    ;
            
    var
      rx = xVec - xVec.mean, ry = yVec - yVec.mean,
      beta1 = (rx * ry).sum / (rx ^ 2).sum,
      beta0 = yVec.mean - beta1 * xVec.mean
    ;
            
    f = (x) => beta0 + beta1 * x;
    var 
      n = xVec.count,
      predictions = xVec.transform(f),
      residuals = yVec - predictions,
      rss = ((yVec - predictions) ^ 2).sum,
      tss = ((yVec - yVec.mean) ^2).sum,
      rse = Math.sqrt(rss / (Math.max(n - 2, 1))),
      int_se = rse * Math.sqrt((1 / n + xVec.mean * xVec.mean / (rx ^ 2).sum)),
      gra_se = rse * Math.sqrt(1 / (rx ^ 2).sum),
      gra_t = beta1 / gra_se
    ;
    
    info = {
      "predictions": predictions,
      "residuals": residuals,
      "RSS": rss,
      "RSE": rse,
      "Intercept": beta0,
      "Intercept-SE": int_se,
      "Gradient": beta1,
      "Gradient-SE": gra_se,
      "Gradient-t": gra_t,
      "R-Squared": 1 - rss / tss
    };
  }
  
  /// A summary of the model as markdown.
  String get summary {
    var sb = new StringBuffer();
    sb.write("|Statistic|Value|\n|:-------:|:---:|\n");
    info.forEach((stat, val) {
      if (val is num) 
        sb.write("|$stat|${(val * 10000).round() / 10000}|\n");
    });
    return sb.toString();
  }
}
