part of fisher;

/// An indexable numeric vector.
class NumericVector extends _Vector {
  @override List<num> _data;
  
/// The instances stored in the vector.
  @override List<num> get data => new List<num>.from(_data);
  
  Map _cache = {};
  void clearCache() {
    _cache.clear();
  }
  
  NumericVector({var data, String name}) {
    this.name = name == null ? DEFAULT_VECTOR_NAME : name;
    _data = _extractNumList(data);
  }
  
  
  
/// Returns a vector containing the elements transformed under [f].
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     var b = a.transform((x) => 2 * x - 1);
///
/// ...
  NumericVector transform(Function f) => 
    new NumericVector(data: this.map(f), name: name);
  
/// Returns a Vector containing only the elements for which [predicate] 
/// returns true.
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     var b = a.takeWhere((x) => x % 2 == 0);
/// 
/// ...
  NumericVector takeWhere(Function predicate) => 
    new NumericVector(data: this.where(predicate), name: name);
  
  operator []= (var index, var value) {
    List<num>
      indexList = _extractNumList(index),
      valueList = _extractNumList(value)
    ;
    int length = Math.max(indexList.length, valueList.length);
    for (int i = 0; i < length; i++) 
      _data[indexList[i % indexList.length] % count] = 
        valueList[i % valueList.length]; 
  }
  
/// The first instance in the vector.
  num get first => _data.first;

/// The last instance in the vector.
  num get last => _data.last;


/// An interpretation of the [NumericVector] as a [DataMatrix].
/// 
/// The vector is interpreted as a column-matrix. Use [transpose]
/// for a row-matrix interpretation.
  DataMatrix get asDataMatrix => new DataMatrix([this]);

/// An interpretation of the [NumericVector] as a [DataMatrix].
/// 
/// The vector is interpreted as a row-matrix. Use [asDataMatrix]
/// for a column-matrix interpretation.
  DataMatrix get transpose => new DataMatrix(
    new List<NumericVector>.generate(
      count, 
      (int i) => new NumericVector(data: [this[i]])
    )
  );
  
  _elementWiseOperation(var x, Function f) =>
    x is num ? _elementWiseOperation([x], f) :
      x is NumericVector ? _elementWiseOperation(x.data, f) :     
        new NumericVector(
          data: new List<num>.generate(
            count, 
            (int i) => f(_data[i], x[i % x.length])
          ),
          name: name
        );
  
/// Performs element-wise addition.
/// 
/// Expects [x] to be of type [num], [List] or [NumericVector]. Elements
/// are reused if [count]s do not match.
  operator + (var x) => 
    _elementWiseOperation(x, (num a, num b) => a + b);
  
/// Performs element-wise negation.
  operator - () => 
    transform((x) => -x);
  
/// Performs element-wise subtraction.
/// 
/// Expects [x] to be of type [num], [List] or [NumericVector]. Elements
/// are reused if [count]s do not match.
  operator - (var x) => 
    this + (-x);
  
/// Performs element-wise- or matrix-multiplication.
/// 
/// If [x] is of type [num], [List] or [NumericVector], element-wise
/// multiplication is performed with reuse if [count]s do not match.
/// 
/// For the dot-product as a [num], use [dot].
/// 
/// If [x] is of type [DataMatrix] then matrix multiplication between
/// the vector's [transpose] and [x] is performed.
  operator * (var x) => 
    x is DataMatrix ? transpose * x :
      _elementWiseOperation(x, (num a, num b) => a * b);
  
/// Performs element-wise division.
/// 
/// Expects [x] to be of type [num], [List] or [NumericVector]. Elements
/// are reused if [count]s do not match.
  operator / (var x) => 
    _elementWiseOperation(x, (num a, num b) => a / b);
  
/// Performs element-wise integer division.
/// 
/// Expects [x] to be of type [num], [List] or [NumericVector]. Elements
/// are reused if [count]s do not match.
  operator ~/ (var x) => 
    _elementWiseOperation(x, (num a, num b) => a ~/ b);
  
/// Performs element-wise modulus division.
/// 
/// Expects [x] to be of type [num], [List] or [NumericVector]. Elements
/// are reused if [count]s do not match.
  operator % (var x) => 
    _elementWiseOperation(x, (num a, num b) => a % b);
  
/// Performs element-wise power calculations.
/// 
/// Expects [x] to be of type [num], [List] or [NumericVector]. Elements
/// are reused if [count]s do not match.
  operator ^ (var x) => 
    _elementWiseOperation(x, (num a, num b) => Math.pow(a, b));
   
  BooleanVector _elementWiseConditional(var x, Function f) =>
    x is num ? _elementWiseConditional([x], f) :
      x is NumericVector ? _elementWiseConditional(x.data, f) :
        new BooleanVector(
          data: new List<bool>.generate(
            count,
            (int i) => f(_data[i], x[i % x.length])
          ),
          name: name
        );

/// Returns a [BooleanVector] with the results of an element-wise
/// less-than test.
  operator < (var x) => 
    _elementWiseConditional(x, (num a, num b) => a < b);

/// Returns a [BooleanVector] with the results of an element-wise
/// less-than-or-equals test.
  operator <= (var x) => 
    _elementWiseConditional(x, (num a, num b) => a <= b);

/// Returns a [BooleanVector] with the results of an element-wise
/// greater-than test.  
  operator > (var x) => 
    _elementWiseConditional(x, (num a, num b) => a > b);
  
/// Returns a [BooleanVector] with the results of an element-wise
/// greater-than-or-equals test.
  operator >= (var x) => 
    _elementWiseConditional(x, (num a, num b) => a >= b);

/// Returns a [BooleanVector] with the results of an element-wise
/// equality test.
  @override 
  operator == (var x) => 
    _elementWiseConditional(x, (num a, num b) => a == b);
  
/// Returns the dot product with [x].
/// 
/// Expects [x] to be a [NumericVector], a [List] or a [num]. Elements
/// are reused if [count]s do not match.
  dot(var x) {
    var helper = (List<num> a) => 
      (
        new List<num>.generate(
          count,
          (int i) => _data[i] * a[i % a.length]
        )
      ).fold(
        0, 
        (num a, num b) => a + b
      )
    ;
    if (x is num) return helper([x]);
    else if (x is NumericVector) return helper(x.data);
    else return helper(x);
  } 
  
  _checkCache(String statistic, Function fall) =>
    _cache.containsKey(statistic) ? _cache[statistic] :
      _cache[statistic] = fall();
  
/// The minimum of the elements.
  num get min => _checkCache(
    "min", () => _data.reduce(Math.min));
  
/// The maximum of the elements.
  num get max => _checkCache(
    "max", () => _data.reduce(Math.max));
  
/// The range of the elements.
  num get range => max - min;
  
/// The sum of the elements.
  num get sum => _checkCache(
    "sum", () =>_data.fold(0, (num a, num b) => a + b));
  
/// The mean of the elements.
  num get mean => _checkCache(
    "mean", () => sum / count);
  
/// The magnitude of the vector.
  num get magnitude => _checkCache(
    "magnitude", () => Math.sqrt(dot(this)));

/// Returns the [k]th central moment of the data.
  num centralMoment(int k) => 
    transform((num x) => Math.pow(x - mean, k)).mean;

/// Returns the [k]th standardised central moment of the data.
  num stdCentralMoment(int k) =>
    centralMoment(k) / Math.pow(stdDev, k);
  
/// The variance in the elements.
  num get variance => _checkCache(
    "variance", () => centralMoment(2) 
      * count / (count == 1 ? 1 : (count - 1)));
  
/// The skewness in the elements (defined as the third standardised 
/// central moment).
  num get skewness => _checkCache(
    "skewness", () => stdCentralMoment(3));
  
/// The kurtosis in the elements (defined as three less than the 
/// fourth standardised central moment).
  num get kurtosis => _checkCache(
    "kurtosis", () => stdCentralMoment(4)) - 3;
  
/// The standard deviation in the elements.
  num get stdDev => _checkCache(
    "stdDev", 
    () => Math.sqrt(variance)
  );
  
/// Returns the quantile bounding proportion [p] of the data.
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     print(a.quantile(0.99));
///     
/// ...
  num quantile(num p) {
    if (_data.length == 0 || p < 0 || p > 1) return null;
    else if (_data.length == 1) return _data.first;
    else {
      var sorted = new List<num>.from(_data)..sort();
      int index = ((count - 1) * p).floor();
      num interpolate = (count - 1) * p - index;
      return sorted[index] + (sorted[index + 1] - sorted[index]) * interpolate;
    }
  }
  
/// The median of the elements.
  num get median => _checkCache(
    "median",
    () => quantile(0.5)
  );
  
/// The lower quartile of the elements.
  num get q1 => _checkCache(
    "q1",
    () => quantile(0.25)
  );
  
/// The upper quartile of the elements.
  num get q3 => _checkCache(
    "q3",
    () => quantile(0.75)
  );
  
/// The inter-quartile range of the elements.
  num get iqr => q3 - q1;
  
/// The z-scores of the elements as a [NumericVector].
  NumericVector get zScores => transform((num x) => (x - mean) / stdDev);
  
/// The mean absolute deviation within the elements.
  num get mad => transform((num x) => (x - mean).abs()).mean;
  
/// A bootstrap sample of the elements, that is, a random resampling 
/// with replacement.
/// 
/// ...
/// Example:
///
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     print(a.bootstrapSample);
///     print(a.bootstrapSample);
///     print(a.bootstrapSample);
///     
/// ...
  NumericVector get bootstrapSample => sample(count);
  
/// Returns a sample of size [n] of the elements chosen at random.
/// 
/// The argument [withReplacement] determines whether replacement 
/// is used and has a default value of true.
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     print(a.sample(3));
///     print(a.sample(4, withReplacement: false);
///     
/// ...
  NumericVector sample(int n, {bool withReplacement: true}) {
    if (withReplacement) {
      return new NumericVector(
        data: new List<num>.generate(
          n, 
          (_) => _data[_rand.nextInt(count)]
        ),
        name: name
      );
    } else {
      if (n > count) return null;
      else {
        var available = new List<int>.generate(count, (int i) => i);
        return new NumericVector(
          data: new List<num>.generate(
            n,
            (_) {
              int i = available[_rand.nextInt(available.length)];
              available.remove(i);
              return i;
            }
          ),
          name: name
        );
      }
    }
  }
  
  num _getStatistic(String stat) => ({
    "kurtosis": kurtosis,
    "mad": mad,
    "magnitude": magnitude,
    "mean": mean,
    "median": median,
    "q1": q1,
    "q3": q3,
    "skewness": skewness,
    "stdDev": stdDev,
    "sum": sum,
    "variance": variance
  })[stat];
  
/// Returns an estimated p-value associated with a statistic value.
/// 
/// Expects [parameter] to be a [String] representation of any
/// characteristic inherent to [NumericVector] and [atLeastAs] to be
/// one of "extreme", "great" or "small".
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     print(a.bootstrapPValue(
///       parameter: "mean", 
///       atLeastAs: "extreme",
///       asValue: 8.5
///     ));
///     
/// ...
  num bootstrapPValue({
    String parameter: "mean",
    String atLeastAs: "extreme",
    num asValue,
    int simulations: 10000
  }) {
    var statistics = new NumericVector(
      data: new List<num>.generate(
        simulations, 
        (_) => bootstrapSample._getStatistic(parameter.toLowerCase())
      ),
      name: name
    );
    
    num 
      mu = statistics.mean,
      statistic = asValue,
      d = (statistic - mu).abs()
    ;
    
    switch(atLeastAs.toLowerCase()) {
      case "extreme": return 
        ((statistics > mu + d) | (statistics < mu - d)).sum / simulations;
      case "great": return
        (statistics > mu + d).sum / simulations;
      case "small": return
        (statistics < mu - d).sum / simulations;
      default: return null;
    }
  }
  
/// Returns an estimated confidence interval associated with a statistic.
/// 
/// Expects [parameter] to be a [String] representation of any
/// characteristic inherent to [NumericVector]. The argument [p] is 
/// the estimated probability to be captured by the interval.
/// 
/// The interval is returned as a [List] containing the lower and upper
/// bounds.
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([2, 4, 5, 7, 8, 9]);
///     print(a.bootstrapCI(
///       parameter: "mean", 
///       p: 0.95
///     ));
///     
/// ...
  List<num> bootstrapCI({
    String parameter: "mean",
    num p: 0.95,
    int simulations: 10000
  }) {
      var statistics = new NumericVector(
        data: new List<num>.generate(
          simulations, 
          (_) => bootstrapSample._getStatistic(parameter)
        ),
        name: name
      );
      
      return [
        statistics.quantile((1 - p) / 2), 
        statistics.quantile(p + (1 - p) / 2)
      ];
    }
  
/// Returns a [NumericVector] with the data sorted numerically.
/// 
/// The data is arranged in descending order if [descending] is true
/// or in ascending order otherwise. The vector isn't actually changed
/// 
/// ...
/// Example:
/// 
///     var a = new NumericVector([1, 0, 2, 9, 3, 8, 4, 7, 5, 6]);
///     print(a.sorted());
///     
/// ...
  NumericVector sorted({bool descending: false}) =>
    new NumericVector(
      data: data..sort(
        (num a, num b) => descending ? 
          b.compareTo(a) : 
            a.compareTo(b)
      )
    );

/// Returns a [List] of indices of the sorted data.
/// 
/// The vector itself is unchanged. Compare with [sort].
  List<int> sortedIndices({bool descending: false}) => 
      sorted(descending: descending).map((num x) => _data.indexOf(x));

}